from . import restClient,query,file,utils,Sobjects,traceFlag,elementParser

import colorama
import sys,time,os
import ansi2html,re
import threading,traceback
from queue import Queue

def printLogRecords(loguser=None,limit=50,whereClause=None):
    logUserId = get_loguser_id(loguser) if loguser != None else None
    if loguser != None:
        print(f'Logs for user {loguser}:')
    logs = get_apexLog_records(logUserId,limit=limit,whereClause=whereClause)
    logs = utils.deleteNulls(logs,systemFields=False)
    logs1 = []
    for log in logs:
        log['LastModifiedDate'] = log['LastModifiedDate'].split('.')[0]
        log['StartTime'] = log['StartTime'].split('.')[0]
        log['LogUserId'] =  f"{log['LogUserId']} ({get_username_and_cache(log['LogUserId'])})"

        logs1.append(log)

    utils.printFormated(logs1,rename="LogLength%Len:DurationMilliseconds%ms:Application%App")
    return logs

def get_apexLog_records(logUserId=None,limit=50,whereClause=None):
    where = f" where {whereClause} " if whereClause != None else ''
    where = f" where logUserId='{logUserId}' " if logUserId is not None else where

    call = query.query(f"Select Id,LogUserId,LogLength,LastModifiedDate,Request,Operation,Application,Status,DurationMilliseconds,StartTime,Location,RequestIdentifier FROM ApexLog  {where} order by LastModifiedDate desc limit {limit}")
    return call

def get_apexLog_record_and_body(logId):
    logRecords = query.queryRecords(f"Select fields(all) FROM ApexLog where Id ='{logId}' limit 1")

    if logRecords == None or len(logRecords)==0:
        utils.raiseException(errorCode='NO_LOG',error=f'The requested log <{logId}> cannot be found in the Server.',other=f"No record in ApexLogwith Id {logId}")    
    logRecord = logRecords[0]

    action = f"/services/data/v56.0/sobjects/ApexLog/{logId}/Body/"
    logbody = restClient.callAPI(action)
    return logRecord,logbody

userCache = {}
def get_username_and_cache(Id):
    username_query = f"select Username from User where Id='{Id}'"
    if username_query not in userCache: userCache[username_query] = query.queryField(username_query) 
    return userCache[username_query]

def apexLog_record_to_string(logRecord):
    log = logRecord
    username = get_username_and_cache(log['LogUserId'])

    logLine = f"""{utils.CFAINT}LOGDATA:    Id: {log['Id']}   LogUserId: {log['LogUserId']} {utils.CGREEN}({username}){utils.CEND}{utils.CFAINT}    Request: {log['Request']}  Operation: {utils.CGREEN}{log['Operation']}{utils.CEND}{utils.CFAINT}    lenght: {log['LogLength']}    duration:  {utils.CGREEN}{log['DurationMilliseconds']} {utils.CEND} 
 {utils.CFAINT}LOGDATA:      startTime: {log['StartTime']}    app: {log['Application']}      status: {log['Status']}     location: {log['Location']}     requestIdentifier: {log['RequestIdentifier']}{utils.CEND}
    """     
    return logLine

def do_parse_storage(pc,search_dir=None):  
    if pc['store_logId'] != None:
       # pc['logId'] = pc['store_logId']
        pc['filepath'] = f"{restClient.logFolder()}{pc['store_logId']}.log"

        do_parse_logId(pc)
        return

    if search_dir==None:
        search_dir = restClient.logFolder()

    os.chdir(search_dir)
    files = filter(os.path.isfile, os.listdir(search_dir))
    files = [os.path.join(search_dir, f) for f in files] # add path to each file
    files.sort(key=lambda x: os.path.getmtime(x))
    log_files = [f for f in files if '.log' in f]
    fileNames = [os.path.basename(f) for f in log_files]

    print(f"Files to be parsed in the store {len(log_files)}")
    file_dates = []

    print(f"Ordering files by date...")

    for log_file in log_files:
        body = get_apexLog_body_from_file(log_file)
        for line in body.splitlines():
            if '|' in line:
                _time = line.split(' ')[0]
                file_dates.append({
                    'time':_time,
                    'file':log_file
                })
                break

    newlist = sorted(file_dates, key=lambda d: d['time'])

    sorted_log_file = [d['file'] for d in newlist]

    print(f"Ordered.")

    try:
        parse_apexlogs_by_Ids_or_filepaths(pc,logIds=None,filepaths=sorted_log_file,printProgress=True,printNum=False,threads=0)

    except KeyboardInterrupt:
        print('Interrupted')
    
    print_parsing_results(pc)

   # print(frequency)

def do_parse_tail(pc):
    def auto_renew_traceFlag(traceFlagId):
        try:
            while True:
                traceFlag.update_trace_flag_incli(traceFlagId,minutes=5)
                time.sleep(10)
        except Exception as e:
            print()

    def deleteRecords(delete_queue):
        def divide_chunks(l, n):
            for i in range(0, len(l), n):
                yield l[i:i + n]
            
        while True:
            logIds = delete_queue.get()

            try :
                logIdsList= list(divide_chunks(logIds,200))
                for l in logIdsList:
                    res = Sobjects.deleteMultiple('ApexLog',l)
                restClient.glog().debug(f"deleted records {logIds}")
                delete_queue.task_done()
            except utils.InCliError as e:
                if e.args[0]['errorCode'] != 'NO_LOG':
                    utils.printException(e)
            except Exception as e:
                print(logIds)
                print(e)

    timefield = "LastModifiedDate"

    logRecords = query.queryRecords(f"Select fields(all) FROM ApexLog order by {timefield} desc limit 1")
    time0 = logRecords[0][timefield] if len(logRecords) > 0 else None
    timez = time0.split('.')[0] + "Z" if time0 != None else '2000-12-12T17:19:35Z'

    delete_queue= None
    restClient.glog().debug(f"deleteLogs-->{pc['deleteLogs']}")

    if (pc['deleteLogs'] or pc['auto']) and pc['loguser'] == None:  
        pc['loguser'] = f"username:{pc['connection']['Username']}"
    logUserId = get_loguser_id(pc['loguser']) if pc['loguser'] != None else None

    if pc['deleteLogs']==True:       
        restClient.glog().debug("Starting delete queue")
        delete_queue = Queue(maxsize=0)
        for x in range(0,1):
            threading.Thread(target=deleteRecords,args=(delete_queue,), daemon=True).start()
        restClient.glog().info(f"Auto delete Apexlog records for user {pc['loguser']} {logUserId} set to Auto")

    if pc['auto']:
       # logUserId = get_loguser_id(f"username:{pc['loguser']}") if logUserId == None else logUserId
        traceFlagId = traceFlag.set_incli_traceFlag_for_user(f"Id:{logUserId}")
        restClient.glog().info(f"TraceFlag for user {pc['loguser']} {logUserId} set to Auto")

        threading.Thread(target=auto_renew_traceFlag,args=(traceFlagId,), daemon=True).start()

    try:
        waitingPrinted = False
        procesed = []
        greater = True
        while (True):
            if greater:    where = f" {timefield} > {timez} "
            else:          where = f" {timefield} >= {timez} "

            where = f" {pc['whereClause']} and {where}" if pc['whereClause'] is not None else where
            where = f" logUserId='{logUserId}'and {where} " if logUserId is not None else where

            fields = "Id,LogUserId,LogLength,LastModifiedDate,Request,Operation,Application,Status,DurationMilliseconds,StartTime,Location,RequestIdentifier,SystemModstamp"
            logRecords = query.queryRecords(f"Select {fields} FROM ApexLog where {where} order by {timefield} asc")
            if len(logRecords) > 0:
                waitingPrinted = False

                prelogIds = [record['Id'] for record in logRecords]
                logIds = [id for id in prelogIds if id not in procesed]
                if len(logIds) == 0:
                    greater = True
                    continue
                greater = False
                procesed.extend(logIds)

                parse_apexlogs_by_Ids_or_filepaths(logIds=logIds,pc=pc,raiseKeyBoardInterrupt=True)

                time0 = logRecords[-1][timefield]
                timez = time0.split('.')[0] + "Z"

                if delete_queue!=None:
                    delete_queue.put(logIds)
                    restClient.glog().debug(f"{logIds} into queue...")

            elif  waitingPrinted == False:
                print()
                print(f"waiting for debug logs for user {pc['loguser']}")  if pc['loguser'] != None  else print(f"waiting for debug logs ")
                waitingPrinted = True

            time.sleep(2)
    except KeyboardInterrupt as e:
        print()
        print_parsing_results(pc)
        print("Terminating -tail..., cleaning up")
        traceFlag.update_trace_flag_incli(traceFlagId,minutes=1,start=-15)
        if delete_queue != None: 
            while delete_queue.empty()==False:    time.sleep(1)
        print('Terminated')
        return

def do_parse_logs_lastN(pc):
    whereClause = pc['whereClause']
    loguser = pc['loguser']
    lastN = pc['lastN']

    if loguser ==None:
        loguser = restClient.getConfigVar('loguser')

    if loguser == None:
        print(f"{utils.CYELLOW}Getting logs for all users.{utils.CEND}")
    else:
        print(f"{utils.CYELLOW}Getting logs for {loguser}.{utils.CEND}")

    where = f" where {whereClause} " if whereClause is not None else ''
    where = f" where logUserId='{get_loguser_id(loguser)}' " if loguser is not None else where

    if lastN == None: lastN = 1
    q = f"Select Id FROM ApexLog {where} order by LastModifiedDate desc limit {lastN}"
    logIds = query.queryFieldList(q)
    if logIds == None or len(logIds)==0:   utils.raiseException(errorCode='NO_LOG',error=f'No logs can be found. ',other=q)

    parse_apexlogs_by_Ids_or_filepaths(logIds= logIds,pc=pc)
    print_parsing_results(pc)

def do_parse_from_file(parseContext):
    set_apexlog_body_in_pc(parseContext)

  #  parseContext['body'] = file.read(parseContext['filepath'])
    parseContext['operation'] = 'parsefile'
 #   name = os.path.basename(parseContext['filepath']).split('.')[0]
 #   parseContext['logId']=name
    name = os.path.basename(parseContext['filepath'])
    parseContext['logId'] = name.split('.')[0]
    context =  parse_apexlog_body(parseContext)
    print_parsed_lines_to_output(parseContext)
    return context

def do_parse_logId(parseContext):
    set_apexlog_body_in_pc(parseContext)
    context =  parse_apexlog_body(parseContext)
    print_parsed_lines_to_output(parseContext)

    return context

def parse_apexlogs_by_Ids_or_filepaths(pc,logIds=None,filepaths=None,raiseKeyBoardInterrupt=False,printProgress=False,threads=10,printNum=True):
    def read_log_from_org(q):
      while True:
        Id = q.get()
        get_apexLog_body_from_file_otherwise_db(Id)
        restClient.glog().debug(f"Read body for Id {Id}")
        q.task_done()

    if 'total_parsed'       not in pc:   pc['total_parsed'] = 0
    if 'parsed_Id_status'   not in pc:   pc['parsed_Id_status'] = []
    if 'errors' not in pc:   pc['errors'] = []
    if 'queue'  not in pc:   pc['queue'] = None

    if filepaths != None:
        threads = 0

    if threads >0: 
        if pc['queue'] == None:
            pc['queue'] = Queue(maxsize=0)
            for x in range(0,threads):
                threading.Thread(target=read_log_from_org,args=(pc['queue'],), daemon=True).start()
        for logid in logIds:
            pc['queue'].put(logid)
            
    num = 0
    items = filepaths if filepaths != None else logIds
    for num,item in enumerate(items):
        if printProgress:
            sys.stdout.write("\r%d%%" % int(100*num/len(items)))
        try:
            if logIds!= None:
                parsed={ 'logId':item, 'status':'ok' }
                pc['logId'] = item
                pc['filepath'] = None
            else:
                parsed={ 'file':os.path.basename(item), 'status':'ok' }
                pc['filepath'] = item  
                pc['logId'] = None
             
            pc['parsed_Id_status'].append(parsed)
            set_apexlog_body_in_pc(pc)
            parse_apexlog_body(pc)
            print_parsed_lines_to_output(pc)
            if printNum:    print( pc['total_parsed']+num+1)
            if pc['context']['exception'] == True:    parsed['status'] = pc['context']['exception_msg'][0:200]

        except KeyboardInterrupt:
            if raiseKeyBoardInterrupt:        raise
            break
        except utils.InCliError as e:
             parsed['status'] = f"Parse error: {e.args[0]['errorCode']}  "
             utils.printException(e)
             pc['errors'].append(e)
        except Exception as e:
            parsed['status'] = f"Unknown: {e}"
            pc['errors'].append(e)
            print(traceback.format_exc())

    pc['total_parsed'] = pc['total_parsed'] + num + 1
    
def print_parsing_results(pc):
    print()

    if 'parsed_Id_status' not in pc:
        print("No files parsed.")
        return 
    parsed = pc['parsed_Id_status']
    errors = pc['errors']

    print(f"{pc['total_parsed']} logs parsed")
    parsed = [par for par in parsed if par['status']!='ok']

    if len(parsed) == 0:  print("No errors.")
    if len(parsed)>0:
        utils.printFormated(parsed)
        errors = list({error.args[0]['errorCode']:error for error in errors}.values())
        for error in errors:    utils.printException(error)  

def get_loguser_id(loguser):
    id = Sobjects.IdF('User',loguser)
    return id if id!= None else utils.raiseException('QUERY',f"User with field {loguser} does not exist in the User Object.") 

def get_apexLog_body_from_file(filename):
    if file.exists(filename):
        return file.read(filename)

    return None

def set_apexlog_body_in_pc(pc):
    """if pc['filepath'] defined, reads from the file specified. Otherwiese logId needs to be set. 
    """
    if 'filepath' in pc and pc['filepath'] != None:
        pc['logRecord'] = None
        pc['body'] = get_apexLog_body_from_file(pc['filepath'])
    else:
        pc['logRecord'],pc['body'],pc['filepath'] = get_apexLog_body_from_file_otherwise_db(pc['logId'])
    if pc['body'] == None :   utils.raiseException(errorCode='NO_LOG',error=f'The requested log <{pc["logId"]}> cannot be found. ')
    if len(pc['body'])==0:    utils.raiseException(errorCode='NO_LOG',error=f'The body for the requested log <{pc["logId"]}> is empty. ')
    
def get_apexLog_body_from_file_otherwise_db(logId):
    """Gets the log body for the provided logId from file (if exists) otherwise from the Org"""
    filename = f"{restClient.logFolder()}{logId}.log"

    if file.exists(filename) == True:
        body = get_apexLog_body_from_file(filename)

        if body == None or len(body)==0:
            print("The file seems corrupted. Getting log from server.")
            file.delete(filename)
            return get_apexLog_body_from_file_otherwise_db(logId)
        return None,body,filename
    else:
        logRecord,body = get_apexLog_record_and_body(logId) 
        body = apexLog_record_to_string(logRecord) + body  
        save_to_store(logId,body)
        return logRecord,body,filename

def save_to_store(logId,body):
    filename = f"{restClient.logFolder()}{logId}.log"
    file.write(filename,body) 

def delta(obj,field):
    return obj[field][1] - obj[field][0] if len(obj[field]) > 1 else 0

def createContext(lines):
    context = {
        'totalQueries' : 0,
        'timeZero':0,
        'ident':0,
        'DEF:SOQL queries' : 0,
        'DEF:CPU time' : 0,
        'CMT:SOQL queries' : 0,
        'CMT:CPU time' : 0,
        "exception":False
    }
    context['totalQueries'] = 0
    context['timeZero'] = 0
    context['ident'] = 0
    context['DEF:SOQL queries'] = 0
    context['DEF:CPU time']=0
    context['CMT:SOQL queries'] = 0
    context['CMT:CPU time']=0
    context['exception'] = False
    context['file_exception'] = False
    context['previousElapsedTime'] = 0
    context['previousCPUTime'] = 0
    context['previousIsLimit'] = False
    context['prevTimes'] = {
        0:[0,0]
    }
    context['prevLevel'] = 0
    context['firstLineIn'] = True
    context['firstLineOut'] = True

    context['parsedLines'] = []
    context['openItemsList'] = []

    context['lines'] = lines

    return context

frequency = {}
def parse_apexlog_body(pc):
    if pc['body'] == None :  utils.raiseException(errorCode='NO_LOG',error=f'The requested log <{pc["logId"]}> cannot be found. ')
    if len(pc['body'])==0:    utils.raiseException(errorCode='NO_LOG',error=f'The body for the requested log <{pc["logId"]}> is empty. ')

    try:
        context = createContext(pc['body'].splitlines())
        pc['context'] = context

        for num,line in enumerate(context['lines']):
            if context['firstLineIn'] == True:
                if 'APEX_CODE' in line:
                    context['firstLineIn'] = False
                    levels = line.strip().split(' ')[1].replace(',','=').replace(';','  ')
                    levels = f"{utils.CFAINT}{levels}{utils.CEND}"
                    obj = {  'type':'LOGDATA', 'output':levels  }
                    context['parsedLines'].append(obj)

                    continue      
                else:
                    obj = {    'type':'LOGDATA',   'output':line  }
                    context['parsedLines'].append(obj)
                    continue

            chunks = line.split('|')
            if len(chunks)<2:
                continue
            if len(chunks[0])<10:
                continue
            if len(chunks[1])>30:
                continue

            context['chunks'] = chunks
            context['chunks_lenght'] = len(chunks)
            context['line'] = line
            context['line_index'] = num

            if len(chunks)>1 and chunks[1] in ['SYSTEM_MODE_ENTER','SYSTEM_MODE_EXIT','HEAP_ALLOCATE','STATEMENT_EXECUTE','VARIABLE_SCOPE_BEGIN','HEAP_ALLOCATE','SYSTEM_METHOD_ENTRY','SYSTEM_METHOD_EXIT','SOQL_EXECUTE_EXPLAIN','ENTERING_MANAGED_PKG','SYSTEM_CONSTRUCTOR_ENTRY','SYSTEM_CONSTRUCTOR_EXIT']:    continue

            if len(chunks)>1 and chunks[1] in ['VALIDATION_RULE','VALIDATION_FORMULA','VALIDATION_PASS','WF_RULE_FILTER','WF_RULE_EVAL_VALUE','STATIC_VARIABLE_LIST','CUMULATIVE_LIMIT_USAGE_END','CUMULATIVE_LIMIT_USAGE','FLOW_CREATE_INTERVIEW_BEGIN','FLOW_CREATE_INTERVIEW_END','TOTAL_EMAIL_RECIPIENTS_QUEUED','CUMULATIVE_PROFILING_BEGIN','CUMULATIVE_PROFILING','CUMULATIVE_PROFILING_END','EXECUTION_STARTED','EXECUTION_FINISHED'] : continue

            if elementParser.parseVariableAssigment(context):      continue
            if elementParser.parseMethod(context):            continue
            if elementParser.parseSOQL(context)==True:        continue
            if elementParser.parseLimits(context):     continue
            if elementParser.parseUserDebug(context):  continue
            if elementParser.parseUserInfo(context):   continue
            if elementParser.parseExceptionThrown(context):   continue
            if elementParser.parseDML(context):               continue
            if elementParser.parseConstructor(context):       continue
            if elementParser.parseCodeUnit(context):          continue
            if elementParser.parseConstructor(context):       continue
            if elementParser.parseNamedCredentials(context): continue
            if elementParser.parseCallOutResponse(context): continue
            if elementParser.parseWfRule(context) : continue
            if elementParser.parseFlow(context) : continue
            
        if len(context['openItemsList']) > 0:
            a=1
        elementParser.appendEnd(context)

        return context

    except KeyboardInterrupt as e:
        print(f"Parsing for logI {pc['logId']} interrupted.")
        raise e
    except Exception as e:
        print(f"Exception while parsing for logI {pc['logId']} ")
        raise e


def print_parsed_lines_to_output(pc):
    def escape_ansi(line):
        ansi_escape =re.compile(r'(\x9B|\x1B\[)[0-?]*[ -\/]*[@-~]')
        return ansi_escape.sub('', line)

    if 1==1:
        pc['context']['parsedLines'] = processRepetition(pc['context']['parsedLines'])

    logId = pc['logId']
    toFile= pc['writeToFile'] if 'writeToFile' in pc else False

    if toFile == True:
        filename = f"{restClient.logFolder()}{logId}_ansi.txt"

        original_stdout = sys.stdout
        with open(filename, 'w') as f:
            sys.stdout = f 
            print_parsed_lines(pc)
            sys.stdout = original_stdout 

        data = file.read(filename)
        converter = ansi2html.Ansi2HTMLConverter()
     #   converter.
        html = converter.convert(data)
        html = html.replace('background-color: #000000;','background-color: #020b3b;')
        html = html.replace('aa5500','ecec16')
        html = html.replace('F5F1DE','8e8383')
        html = html.replace('AAAAAA','FFFFFF')
        html = html.replace('000316','3c3e48')
        html = html.replace('aa0000','e41717')

        filename = f"{restClient.logFolder()}{logId}.html"
        file.write(filename,html)
        print(f"Html file: {filename}")
        clean = escape_ansi(data)
        filename = f"{restClient.logFolder()}{logId}.txt"
        file.write(filename,clean)  
        print(f"Txt file: {filename}")
 
    else:
        colorama.just_fix_windows_console()
        print_parsed_lines(pc)

def print_parsed_lines(pc):
    logId = pc['logId'] if 'logId' in pc else None
    context = pc['context']
    print_only_errors = pc['print_only_errors'] if 'print_only_errors' in pc else False
    if context['exception'] == False and print_only_errors == True: return 
    print('___________________________________________________________________________________________________________________________________________________________________')
    if logId != None:  print(logId)
    print()
    if logId != None:  print(f"{utils.CFAINT}Parsing log Id {logId}    file: {restClient.logFolder()}{logId}.log{utils.CEND}")
    elif pc['filepath'] != None:
            print(f"file: {pc['filepath']}")

    firstLines = True

    pc['new_print'] = True
    for num,parsedLine in enumerate(context['parsedLines']):
        if parsedLine['type'] == 'LOGDATA':
            print(parsedLine['output'])
            continue
        else:
            if firstLines == True:
                firstLines = False
                print()
                print()
        printLimits = pc['printLimits'] if 'printLimits' in pc else False

        if printLimits == False:
            if '*** getCpuTime() ***' in parsedLine['output']:   continue
            if '*** getQueries() ***' in parsedLine['output']:   continue
            if parsedLine['type'] == 'LIMIT':                    continue

        print_parsed_line(pc,parsedLine,num)    
    print()

def isRep(all_loops,parsed_line_index,parsed_line,parsed_lines):
      if parsed_line['type'] not in ['METHOD','WF_CRITERIA','CONSTRUCTOR','DEBUG']:  return False,None

      for loop in all_loops:
        links = len(loop)
        loop_start_position = loop[0]
        link_lenght = loop[1] - loop[0]
        loop_lenght = links * link_lenght

        if loop_start_position <= parsed_line_index <= loop_start_position + loop_lenght - 1:
            if loop_start_position <= parsed_line_index <= loop_start_position + link_lenght-1:

                parsed_line['loop'] = links
                parsed_line['is_loop'] = True
                parsed_line['loop_links'] = links
                parsed_line['loop_link_lenght'] = link_lenght
                parsed_line['loop_position'] = parsed_line_index - loop_start_position + 1

                if len(parsed_line['timeStamp']) > 1 and (loop_start_position <= parsed_line_index < loop_start_position + link_lenght):
                    total_method_time = 0
                    total_wait_time = 0
                    for link_index in range(0,links):
                        brother_index = parsed_line_index + link_index * link_lenght
                        brother = parsed_lines[brother_index]
                        brother_method_time = brother['timeStamp'][1] - brother['timeStamp'][0]
                        total_method_time = total_method_time + brother_method_time

                        if link_index !=0:
                            brother_previous_line = parsed_lines[brother_index - 1]

                            if brother_previous_line['ident'] < brother['ident']:
                                wait_time =  brother['timeStamp'][0] - brother_previous_line['timeStamp'][0]
                            else:
                                ts = brother_previous_line['timeStamp'][1] if len(brother_previous_line['timeStamp'])>1 else  brother_previous_line['timeStamp'][0]
                                wait_time =  brother['timeStamp'][0] - ts

                            total_wait_time = total_wait_time + wait_time
                        #    print(f"{brother_previous_line['ident']} {brother_previous_line['output']}  {brother_previous_line['timeStamp'][0]}  {brother_previous_line['timeStamp'][1]} {brother['ident']}  {brother['output']}   {brother['timeStamp'][0]} {brother['timeStamp'][1]}   {wait_time}  {total_wait_time}")
                    parsed_line['totalLoopTime'] = total_method_time
                    parsed_line['totalLoopWait'] = total_wait_time
    
                #for the last line in the chain, set time_exit to the very last in the loop
                if len(parsed_line['timeStamp']) > 1 : 
                    parsed_line['timeStamp'][1] = parsed_lines[ parsed_line_index + (links-1) * link_lenght ]['timeStamp'][1]
                return True,True

            else: return True,False

      return False,False

def processRepetition1(parsed_lines):

    for parsed_line in parsed_lines:
        if 'output' not in parsed_line:  
            print()

    all_loops = utils.get_all_loops(parsed_lines,"output")

    new_parsed_lines = []
    for parsedLineNum,parsedLine in enumerate(parsed_lines):
        if 'apexline' in parsedLine and parsedLine['apexline'] == "161":
            a=1
        is_loop, is_first_chain = isRep(all_loops,parsedLineNum,parsedLine,parsed_lines)
        if is_loop == True and is_first_chain == False:   continue
        new_parsed_lines.append(parsedLine)

    return new_parsed_lines   

def processRepetition(parsed_lines):

    for parsed_line in parsed_lines:
        if 'output' not in parsed_line:  
            print()

    all_loops = utils.get_all_loops(parsed_lines,"output")

    new_parsed_lines = []
    for parsedLineNum,parsedLine in enumerate(parsed_lines):
        if 'timeStamp' in parsedLine and parsedLine['timeStamp'][0] == 10188648957:
            a=1
            
        is_loop, is_first_chain = isRep(all_loops,parsedLineNum,parsedLine,parsed_lines)
        if is_loop == True and is_first_chain == False:   continue
        new_parsed_lines.append(parsedLine)

    return new_parsed_lines   

def emptyString(context,size,char=' ',ident=None):
 #   str = ''
    if ident is None:   ident = context['ident']
    length = ident * size
    str = " "*length
 #   for x in range(length):   str = str + char  
    return str       

def get_parent_if_last_for_level(parsed_lines,parsed_line_index):
    current = parsed_lines[parsed_line_index]
    for index in range(parsed_line_index+1,len(parsed_lines)):
        if parsed_lines[index]['ident'] == current['ident']:
            return None
        if parsed_lines[index]['ident'] < current['ident']:
            i = 1
            while True:
                parent = parsed_lines[parsed_line_index-i]
                if parent['ident'] < current['ident']:
                    return parent
                i = i + 1


def print_parsed_line(pc,d,parsed_line_index):
    context = pc['context']
    Cinit = utils.CEND

    if pc['new_print']:
        pc['new_print'] = False
        print( f"Log time: {pc['context']['parsedLines'][3]['lines'][0].split(' ')[0]} to {pc['context']['parsedLines'][-1]['lines'][0].split(' ')[0]}")

    if d['type'] == 'LIMITS':
        context['previousIsLimit'] = True
        return

    #levels
    level = d['ident']
    pcLevel = pc['level'] if 'level' in pc else None
    if pcLevel != None:
        if level > int(pc['level']):
            return

    #colors
    element_type = d['type']
    if element_type == 'DEBUG':
        element_type = f"{d['type']}-{d['subType']}"
        Cinit = utils.CRED if d['subType'] == 'ERROR' else utils.CGREEN
    elif element_type == 'VAR_ASSIGN' and d['subType'] == 'EXCEPTION':  Cinit = utils.CRED
    elif element_type == 'VAR_ASSIGN' and d['subType'] != 'EXCEPTION':  return
    elif d['type'] == 'EXCEPTION':  Cinit = utils.CRED
    elif d['type'] == 'SOQL':   Cinit = utils.CCYAN
    elif d['type'] == 'DML':    Cinit =  utils.CYELLOW
    elif d['type'] == 'CODE_UNIT':  Cinit =  utils.CYELLOW

    identation = f"{emptyString(context,1,' ',level)}"

    if 'output' not in d:
        print()
    output_string = d['output']
    
    apex_line_number = d['apexline'] if 'apexline' in d else ''

    cpu_time_entry = int(d['CPUTime'][0])
    cpu_time_exit = int(d['CPUTime'][1]) if len(d['CPUTime']) >1 else ''

    time_stamp_entry = d['timeStamp'][0]
    time_stamp_exit = d['timeStamp'][1] if len(d['timeStamp'])>1 else time_stamp_entry

    soql_total_queries_entry = d['totalQueries'][0]
    soql_total_queries_exit = d['totalQueries'][1] if len(d['SOQLQueries']) >1 else soql_total_queries_entry
    soql_total_queries_delta = soql_total_queries_exit-soql_total_queries_entry

    cpu_time_elapsed = cpu_time_entry - int(context['previousCPUTime'])

    if level == context['prevLevel']:  wait_time = time_stamp_entry - context['prevTimes'][level][1]
    if level >  context['prevLevel']:  wait_time = time_stamp_entry - context['prevTimes'][context['prevLevel']][0]
    if level <  context['prevLevel']:  wait_time = time_stamp_entry - context['prevTimes'][level][1]

    if wait_time <0 and 'is_loop' in d:   
        wait_time = time_stamp_entry - context['prevTimes'][level][0]

    if 'totalLoopWait' in d:
        wait_time = d['totalLoopWait'] + wait_time

    context['prevTimes'][level] = [time_stamp_entry,time_stamp_exit]

    wait_time = f"{wait_time/1000000:.0f}"

    if time_stamp_entry == 475822989:
        a=1
    wait_time_exit = "0"
    parent = get_parent_if_last_for_level( context['parsedLines'],parsed_line_index)
    if parent != None:
        if len(parent['timeStamp']) > 1:
            wait_time_exit = parent['timeStamp'][1] - time_stamp_exit
            wait_time_exit = f"{wait_time_exit/1000000:.0f}"

    context['prevLevel'] = level

    soql_queries_entry = int(d['SOQLQueries'][0])
    soql_queries_exit = d['SOQLQueries'][1] if len(d['SOQLQueries'])>1 else ''
    soql_queries_cmt_entry = d['cmtSOQLQueries'][0]
    soql_queries_cmt_exit = d['cmtSOQLQueries'][1] if len(d['cmtSOQLQueries'])>1 else soql_queries_cmt_entry

    context['previousCPUTime'] = cpu_time_entry
    context['previousElapsedTime']  = d['elapsedTime']

    element_type_color =utils. CYELLOW  if d['type'] in ['SOQL','DML','VAR_ASSIGN'] and level == 0 else ''

    if cpu_time_elapsed == 0 and element_type != 'END':
        cpu_time_elapsed = ''
        cpu_time_entry = ''

    Qmct_estimate = soql_total_queries_entry - soql_queries_entry

    element_time = f"{delta(d,'timeStamp')/1000000:.0f}"
    
    if 'totalLoopTime' in d:  element_time = f"{ d['totalLoopTime']/1000000:.0f}"

    if element_type == 'END':
        soql_queries_exit = d['SOQLQueries'][0]
        cpu_time_entry = d['CPUTime'][0]
        soql_queries_cmt_exit = d['cmtSOQLQueries'][0]

    time_ms = f"{d['elapsedTime']/1000000:.0f}"
    time_ms = f'{int(time_ms):,}'

    if 'is_loop' in d:
        ls = f"    x{d['loop_links']}-{d['loop_position']}"# if d['loop_position'] == 1 else f"             "
        output_string = f"{output_string}   {utils.CYELLOW}{ls}{utils.CEND}"
      #  ls = f"  for:  x{d['loop_links']}  " if d['loop_position'] == 1 else f"             "
      #  output_string = output_string + utils.CYELLOW + f"{ls}{d['loop_position']}" + utils.CEND
    if element_time == "0":  element_time =''
    if soql_total_queries_delta ==0: soql_total_queries_delta = ''
    if soql_total_queries_exit ==0: soql_total_queries_exit = ''
    if Qmct_estimate == 0: Qmct_estimate = ''
    if soql_queries_exit ==0:  soql_queries_exit=''
    if soql_queries_cmt_exit==0: soql_queries_cmt_exit=''
    if soql_total_queries_entry == 0: soql_total_queries_entry = ''
    if wait_time == "0":  wait_time =''
    if wait_time_exit == "0":  wait_time_exit ='' 
    else: wait_time_exit=f"{wait_time_exit}"
    if soql_queries_cmt_entry == 0 : soql_queries_cmt_entry = ''
    if soql_queries_entry == 0 : soql_queries_entry = ''

    time_stamp_entry = utils.CLIGHT_GRAY +utils.CFAINT+ f"{time_stamp_entry:12}" + utils.CEND
    time_stamp_exit = utils.CLIGHT_GRAY +utils.CFAINT+ f"{time_stamp_exit:12}" + utils.CEND

    s = utils.CLIGHT_GRAY +utils.CFAINT+  "|" + utils.CEND

    times_lenght = 10

    if context['firstLineOut'] == True:
       # valname = f"{'wait':>5}     time query  Call Stack"
        valname = f"time      query  Call Stack"

        print(f"{utils.CFAINT}{'time entry':^12}{s}{utils.CFAINT}{'time exit':^12}{s}{'ts':^8}{s}{'CPU':^6}{s}{'cpuD':^6}{s}{'Qt':^3}{s}{'Q':^3}{s}{'Qmp':^3}{s}{'Qe':^3}{s}{'type':^21}{s}{'line':^4}{s}{valname:50}")
        context['firstLineOut'] = False

    if level % 2 == 0:
        et_color = utils.CGREEN + utils.CFAINT
    else:
        et_color = utils.CWHITE + utils.CFAINT

    a=f"{wait_time}·{element_time}·{wait_time_exit}"
    lena = len(a)
    aa = times_lenght-lena
  #  element_time = utils.CGREEN + utils.CFAINT+f"{wait_time}·{utils.CWHITE}{element_time}{utils.CGREEN}{utils.CFAINT}·{wait_time_exit:<{aa}}"
    element_time = utils.CGREEN + f"{wait_time}·{utils.CWHITE}{element_time}{utils.CGREEN}·{wait_time_exit:<{aa}}"

   # output_string = utils.CYELLOW+utils.CFAINT+ f"{wait_time:>5}{wait_time_exit:<4}{identation}"+et_color+f"{element_time:>5}"+utils.CDARK_GRAY+f"{soql_total_queries_delta:>4}" +utils.CEND + Cinit +f"    {output_string}"
    output_string = utils.CYELLOW+utils.CFAINT+ f"{identation}"+f"{element_time:<10}"+utils.CYELLOW+utils.CFAINT+f"{soql_total_queries_delta:>4}" +utils.CEND + Cinit +f"    {output_string}"

    print(f"{time_stamp_entry:12}{s}{time_stamp_exit:12}{s}{time_ms:>8}{s}{cpu_time_entry:>6}{s}{cpu_time_elapsed:6}{s}{soql_total_queries_entry:>3}{s}{utils.CGREEN}{soql_queries_entry:>3}{utils.CEND}{s}{soql_queries_cmt_entry:>3}{s}{utils.CYELLOW}{(Qmct_estimate):>3}{utils.CEND}{s}{element_type_color}{element_type:21}{utils.CEND}{s}{apex_line_number:>4}{s}{output_string:50}"+utils.CEND)






import unittest
from InCli import InCli
from InCli.SFAPI import account,restClient,CPQ

class Test_Account(unittest.TestCase):
    def test_main(self):
        name = 'test1234'
        accountF = f"Name:{name}"

        restClient.init('NOSDEV')
        restClient.setLoggingLevel()
        accountId = account.create_Id(name,recordTypeName='Consumer',checkExists=True)

        dele = account.deleteOrders(accountF)

        cartId = CPQ.createCart(accountF,'Name:B2C Price List','name')
        orders = account.getOrdersId(accountF)
        self.assertTrue(orders!=None)

        dele = account.deleteOrders(accountF)
        orders = account.getOrdersId(accountF)
        self.assertTrue(orders==[])

        accountId2 = account.getId(accountF)
        self.assertTrue(accountId==accountId2)

        acc = account.get(accountId)
        self.assertTrue(accountId==acc['Id'])

        account.delete(accountId)
        acc2 = account.get(accountId)
        self.assertTrue(acc2==None)
        accountId3 = account.getId(f"Name:{name}")
        self.assertTrue(accountId3==None)

        acc3 = account.create(name,recordTypeName='Consumer',checkExists=True)
        accountId4 = account.getId(accountF=f"Name:{name}")
        self.assertTrue(accountId4==acc3['Id'])

        acc4 = account.create(name,recordTypeName='Consumer',checkExists=True)
        self.assertTrue(accountId4==acc4['Id'])
        account.delete(accountId4)

        account.delete(accountId4)


        print()


        



import json
from .SFAPI import Sobjects, query,restClient,jsonFile,utils,objectUtil,debugLogs,digitalCommerceUtil,VlocityErrorLog,ip_attachments,VlocityTrackingEntry
import logging,time
import simplejson,sys
from operator import itemgetter
from inspect import getmembers, isfunction
import traceback
import colorama

#/Users/uormaechea/Documents/Dev/python/InCliLib/InCLipkg/incli/logs/07L3O00000E6dDKUAZ.log

varsFile = 'confIcli.json'
vars = None

def readVars(field=None):
    global vars
    vars = jsonFile.read(varsFile)

    if field != None:
        return vars[field]
    return vars

def setVar(name,value):

    vars = jsonFile.read(varsFile)
    vars[name]=value
    jsonFile.write(varsFile,vars)
    print(simplejson.dumps(vars, indent=4))

def getParam(argv,key,i=1):
    if key not in argv:
        return None

    ind = argv.index(key)
    if (ind+i <= (len(argv)-1)):
        arg = argv[ind+i]
        if arg[0] == '-':
            return None
        return arg
    return None

def listEnvs():
    cons = restClient.getConfigOrgs()
    cons = sorted(cons,key=itemgetter('name'))
    utils.printFormated(cons,fieldsString='name:isSandBox:instance_url:login.username:login.password:login.bearer')
    env = readVars('environment')
    print(f"Current Environment is {env}")

def _queryAndPrint(q,fields=None,systemFields=True,nullFields=False):

    res = query.queryRecords(q)
    if fields != None:
        res = utils.deleteNulls(res,systemFields,nullFields)
        if fields=='all':
            utils.printFormated(res)
        else:
            utils.printFormated(res,fields)
    else:
        res = utils.deleteNulls(res,systemFields,nullFields)
        print(simplejson.dumps(res, indent=4))
    
    print()
    print(f"Null values printed-> {nullFields}  systemFields printed--> {systemFields}")

def option_readme(args):
    if '-h' in args:
        return
    if '-hu' in args:
        return
    
    colorama.just_fix_windows_console()

    print(f"""
    {utils.CYELLOW}Significant updates:{utils.CEND}
     - Changes in version 0.0.61
       Added {utils.CYELLOW}InCli -ipe{utils.CEND} to allow parsing the Integration Procedure Errors.

     - Changes in version 0.0.49
       Added {utils.CYELLOW}-auto{utils.CEND} to the log parsing. This options activates the debug logs automatically for the -loguser especified. If no -loguser is specified it will do it for the user used to authenticate with the org.  
            {utils.CGREEN}InCli -logs -u orgAlias -tail -loguser username:onboarding@nosdti-parceiros.cs109.force.com -deletelogs -auto.{utils.CEND}
            This command will start listening for new logs for user onboarding@nosdti-parceiros.cs109.force.com in DTI and parsing them in real time. 
                -auto will activate the debug logs automatically, so the logs are collected without any configuration in Salesforce. 
                -delete will delete the logs from the server as they are parsed. A local copy for the 
            {utils.CGREEN}InCli -logs -u orgAlias -tail -deletelogs -auto.{utils.CEND}
            This command will start listening for new logs for the user used to authenticate with the org. -auto and -deletelogs as usual.

            * Replace DTI by the correct alias for the org in your sfdx setup. 

    {utils.CYELLOW}Parsed output fields:{utils.CEND}

     - {utils.CGREEN}time entry{utils.CEND}: The entry time to the event, such as METHOD_ENTRY in the debug log. Helpful to search for the line in the debug log
     - {utils.CGREEN}time exit{utils.CEND}:  The exit time for the event, such as METHOD_EXIT in the debug log. Helpful to search for the line in the debug log
     - {utils.CGREEN}ts{utils.CEND}: an esier to read time entry. Time in miliseconds. 
     - {utils.CGREEN}CPU{utils.CEND}: The Salesforce CPU Time. 
            This field is set when there is a CPU time event in the debug log. This are generated by the system and we have no control when they occur. 
            The parser will also look for DEBUG lines containing the CPU time information. The debug line needs to have the following format: "*** getCpuTime() *** Measured: CPUTime"
                Adding this APEX line in the code will increase the accuracy of CPU Time: {utils.CGREEN}System.debug('*** getCpuTime() *** Measured: ' + Limits.getCpuTime());{utils.CEND}
     - {utils.CGREEN}cpuD{utils.CEND}: The CPU time delta from the previous time the CPU Time was parsed. 
     - {utils.CGREEN}Qt{utils.CEND}: Total queries performed executing the APEX program (user apex + managed pacakes apex). This field accurate. 
     - {utils.CGREEN}Q{utils.CEND}: Total queries performed by the user APEX program. This field is almost accurate. When it is not empty is accurate. {utils.CYELLOW}Requires Apex Profiling = Finest.
     - {utils.CGREEN}Qmp{utils.CEND}: Total queries performed by the managed package. We do not have control when they appear. Only when it is updated is accurate.
     - {utils.CGREEN}Qe{utils.CEND}: Estimate queries performed by the managed package. This is an estimate for the manage package queries. Almost accurate, accurate when Q is not empty.
     - {utils.CGREEN}type{utils.CEND}: Type of log event.
     - {utils.CGREEN}line{utils.CEND}: the line in the APEX program. 
     - {utils.CGREEN}wait{utils.CEND}: time elapsed from the previous event. Typically is the time consumed by the APEX program or managed package between events.  
                In some cases there are 2 values x-y. x is the time to the previous event, y is the time after the event. 
     - {utils.CGREEN}time{utils.CEND}: the time consumed executing the method. 
     - {utils.CGREEN}query{utils.CEND}: the total queries consumed bu the method. 
     - {utils.CGREEN}Call Stack{utils.CEND}:
            Some Call Stack string present a  "for: xRepetitions line_number". 
                - Repetitions: how many times the loop is repeated.
                - line_number: the position in the loop
                When there is a loop:
                    - The {utils.CGREEN}time{utils.CEND} field is the total time consumed by the method in the loop. 
                    - The {utils.CGREEN}wait{utils.CEND} field is the total wait time for the method in the loop. 
    
    Times are rounded to the closest milisecond. 
    - APEX Code = Fine is required to show the call stack with methods
    - APEX Profiling = Finest
    - Database = Info
    - Workflow = Info

    """)

def option_vte(args):
    if '-hu' in args: 
        help = f"""
        {utils.CYELLOW}-vte{utils.CEND}    Lists the latest Vlocity Tracking Entries.   {utils.CGREEN}InCli -u orgAlias -vte {utils.CEND}
            {utils.CYELLOW}-limit number{utils.CEND} The number of records to display
        {utils.CYELLOW}-vte a6R7a000001k7t7EAA {utils.CEND} - Displays the JSON for the provided Id.
            {utils.CYELLOW}-file{utils.CEND} Saves the JSON to a file. """ 
        return help

    if '-h' in args: return

    connectionInit(args)
    id = getParam(args,'-vte')
    tofile = True if '-file' in args else False

    if id != None:
        return VlocityTrackingEntry.query_and_print_error_record(id=id,tofile=tofile)

    VlocityTrackingEntry.print_error_list(limit=100)
def option_ipa(args):
    if '-hu' in args: 
        help = f"""
        {utils.CYELLOW}-ipa{utils.CEND}    Lists the latest IP attachments.   {utils.CGREEN}InCli -u orgAlias -ipa {utils.CEND}
            {utils.CYELLOW}-limit number{utils.CEND} The number of records to display
        {utils.CYELLOW}-ipa a2J7a000002UakuEAC {utils.CEND} - Displays the attachment for the provided Id.
            {utils.CYELLOW}-file{utils.CEND} Saves the attachment to a file. """ 
        return help
    if '-h' in args: return

    connectionInit(args)
    ipa_id = getParam(args,'-ipa')
    toFile = True if '-file' in args else False

    if ipa_id != None:
        return ip_attachments.print_attachments(Id=ipa_id,toFile=toFile)

    ip_attachments.print_attachments(limit=100)

def option_ipe(args):
    if '-hu' in args:
        help = f"""
        {utils.CYELLOW}-ipe{utils.CEND} displays the last 50 errors.  {utils.CBLINK}NEW{utils.CEND}
            {utils.CYELLOW}-limit number{utils.CEND} The number of records to display
            {utils.CYELLOW}-last number{utils.CEND} Shows the error Json for the last number of records
        {utils.CYELLOW}-ipe errorId{utils.CEND} Shows the error JSON for the provided Id"""
        return help

    if '-h' in args: return

    connectionInit(args)
    toFile = True if '-file' in args else False
    limit = getParam(args,'-limit')
    last = getParam(args,'-last')
    ipe_id = getParam(args,'-ipe')

    if ipe_id != None:
        return VlocityErrorLog.query_and_print_error_record(ipe_id,tofile=toFile)

    if last != None:
        return VlocityErrorLog.print_error_records(last)

    return VlocityErrorLog.print_error_list(limit)

def option_q(args):
    if '-hu' in args:
        help = f"""
        {utils.CYELLOW}-q "select..."{utils.CEND} query to execute. System Fields, such as LastModifiedDate, and null fields are not returned. 
            {utils.CYELLOW}-null{utils.CEND} - will print fields with null values
            {utils.CYELLOW}-system{utils.CEND} - will print the system fields
            {utils.CYELLOW}-all{utils.CEND} - will print both nulls and system fields
            {utils.CYELLOW}-fields "a:b:c:..."{utils.CEND} --> print in table format the specified fields.
            {utils.CYELLOW}-fields all{utils.CEND} --> print in table format"""
        return help

    if '-h' in args: return

    q = getParam(args,'-q')

    fields = getParam(args,'-fields') if '-fields' in args else None    
    nullFields = True if '-null' in args else False
    systemFields = True if '-system' in args else False
    all = True if '-all' in args else False
    if all == True:
        nullFields = True
        systemFields = True

    connectionInit(args)

    _queryAndPrint(q,fields=fields,systemFields=systemFields,nullFields=nullFields)

def option_cc(args):
    colorama.just_fix_windows_console()

    if '-hu' in args:
        help = f"""
        {utils.CYELLOW}-cc{utils.CEND}       {utils.CGREEN}InCli -u orgAlias -cc {utils.CEND}
            Checks the catalogs. Gets all catalogs, does a getOffers, getOfferDetails. Performs basketwithoutconfig and basket with config, if -basket is set. 
            {utils.CYELLOW}-code{utils.CEND} - Catalogue Code. does it for a single catalog. 
            {utils.CYELLOW}-list{utils.CEND} - List all the catalogues. 
            {utils.CYELLOW}-basket{utils.CEND} - Performs basket operations. """ 

        return help

    if '-h' in args: return

    connectionInit(args)

   # path = getParam(args,'-checkCatalogs')
   # quantity = getParam(args,'-checkCatalogs',2)
    catcatalogueCode = getParam(args,'-code')
    list = True if '-list' in args else False
    basketOps = True if '-basket' in args else False

    account = getParam(args,'-account')

    if list == True:
        digitalCommerceUtil.printCatalogs()
        return

    digitalCommerceUtil.checkOffers(catcatalogueCode=catcatalogueCode,account=account,basketOps=basketOps)

    print()

def option_d(args):
    help = f"""
        {utils.CYELLOW}-d objectName{utils.CEND} --> Describe an object
            {utils.CYELLOW}-d objectName:fieldName{utils.CEND} --> describe a field in the object"""
    if '-hu' in args:
        return help    

    if '-h' in args: return

    connectionInit(args)

    objectField = getParam(args,'-d')
    if objectField == None:
        print(help)
        return

    ofs = objectField.split(':')

    sObjectName = ofs[0]
    fieldName = ofs[1] if len(ofs) > 1 else None

    res = Sobjects.describe(sObjectName)
    if fieldName == None:
        print(simplejson.dumps(res['fields'], indent=4))
    else:
        sibbling = objectUtil.getSiblingWhere(res['fields'],'name',fieldName)['object']

        print(simplejson.dumps(sibbling, indent=4))

def option_l(args):
    if '-hu' in args:
        help = f"""
        {utils.CYELLOW}-l{utils.CEND} --> current org limits consuptions. """
        return help
    if '-h' in args: return
    
    connectionInit(args)
    action = '/services/data/v51.0/limits'
    res = restClient.callAPI(action)
    restClient.checkError()
    records = []
    for key in res.keys():
        record = {
            'Limit':key,
            'Max':res[key]['Max'],
            'Remaining':res[key]['Remaining'],
        }
        record['Percent Remaining'] =  100 *(res[key]['Remaining']/res[key]['Max']) if res[key]['Max']>0 else 0
        record['__color__'] = ''

        if record['Max'] != 0:
            if record['Percent Remaining']<50:
                record['__color__'] = utils.CYELLOW
            if record['Percent Remaining']<25:
                record['__color__'] = utils.CYELLOW
        records.append(record)

    utils.printFormated(records)

def option_o(args):
    if '-hu' in args:
        help = f"""
        {utils.CYELLOW}-o{utils.CEND} --> List all Objects
            {utils.CYELLOW}-name objectName{utils.CEND}  --> get one row from the object
            {utils.CYELLOW}-name objectName:Id{utils.CEND} --> get the row especified by the Id
            {utils.CYELLOW}-like name{utils.CEND} --> where the Name contains the substring name"""

        return help

    if '-h' in args: return
      
    connectionInit(args)

    if '-name' in args:
        obj = getParam(args,'-name')
        limit = getParam(args,'-limit') if '-limit' in args else None    
        l = limit if limit != None else 1
        chunks = obj.split(':')

        if len(chunks) == 1:
            q = f"select fields(all) from {chunks[0]} limit {l}"
        else:
            q = f"select fields(all) from {chunks[0]} where Id='{chunks[1]}'"

        print('pre q:'+q)
        _queryAndPrint(q,systemFields=True,nullFields=True)
        return
        
    like = getParam(args,'-like') if '-like' in args else None    
    count = True if '-count' in args else False   

    objs = Sobjects.listObjects()
    if like is not None:
        if ':' not in like:
            like = f"name:{like}"        
        ls = like.split(':')
        outs = []
        for obj in objs:
      #      print(f"{str(ls[1])} . {str(obj[ls[0]])}")
            if str(ls[1]).lower() in str(obj[ls[0]]).lower():
                outs.append(obj) 
    else:
        outs = objs

    if count==True:
        for out in outs:
            if out['queryable'] == True:
                print("Quering objects row count.")
                print("", end=".")          
                try:
                    c = query.query(f" select count(Id) from {out['name']}",raiseEx=True)
                    out['count'] = c['records'][0]['expr0']
                except Exception as e:
                    out['count'] = 'E'
            else:
                out['count'] = '-'
    
    #print(simplejson.dumps(objs, indent=4))
    utils.printFormated(outs,'name:label:associateParentEntity:associateParentEntity:queryable:count')

def option_default(args):
    help = f"""
        {utils.CYELLOW}-default  {utils.CEND}  --> displays the current default values. 
        {utils.CYELLOW}-default:set key value{utils.CEND} --> defaults the specified key-value. When set as default, if the command does not include the parameter it will pick the default value
        {utils.CYELLOW}-default:del key  {utils.CEND}  --> deletes the default
            {utils.CYELLOW}Defaultable u{utils.CEND}: Can be set to a default value --> {utils.CGREEN}InCli -default:set u collote@acme.com {utils.CEND} or {utils.CGREEN}InCli -default:set u orgAlias {utils.CEND} 
            {utils.CYELLOW}Defaultable loguser{utils.CEND}: Can be set to default. {utils.CGREEN}InCli -default:set loguser "Alias:TheCollote"{utils.CEND}"""

    if '-hu' in args: return

    if '-h' in args:
        return help   

    if "-default:del" in args:
        key = getParam(args,'-default:del',1)
        if key == None:
            restClient.glog().info(f'Key not found in the provided arguments. {args}')
            return 
        restClient.delConfigVar(key)
        return

    if "-default:set" in args: 
        key = getParam(args,'-default:set',1)
        if key == None:
            print(f"{utils.CLIGHT_PURPLE}Parameter to set not specified.{utils.CEND}")
            print(f"{utils.CYELLOW}-default:set key value {utils.CEND}" )
            return        
        value = getParam(args,'-default:set',2)
        if value == None:
            print(f"{utils.CLIGHT_PURPLE}Value to set not specified.{utils.CEND}")
            print(f"{utils.CYELLOW}-default:set key value {utils.CEND}" )
            return        
        if key == None or value==None:
            print(help)
            return

        restClient.setConfigVar(key,value)
        print(f"Default value for {key} is {utils.CYELLOW}{value}{utils.CEND}")

    if "-default" in args: 
        key = 'u' 
        value =restClient.getConfigVar(key)
        if value == None: value = 'Not set'
        print(f"Default value for {key} is {utils.CYELLOW}{value}{utils.CEND}")

        key = 'loguser' 
        value =restClient.getConfigVar(key)
        if value == None: value = 'Not set'
        print(f"Default value for {key} is {utils.CYELLOW}{value}{utils.CEND}")
#def option_history(args):
#    return
#    if '-h' in args:
        help = """
        -history"""
        return help
#    print()
#    print('HISTORY:')
#    vars = jsonFile.read(varsFile)
#    for line in vars['history']:
        print(line)

def option_logs(args):
    help = f"""
        {utils.CYELLOW}-logs Id{utils.CEND} --> parse the log with the provided Id. No modifiers. {utils.CGREEN}InCli -u orgAlias -logs 07L0Q00000N5sMkUAJ"{utils.CEND}

        {utils.CYELLOW}-logs -inputfile {utils.CEND}--> parses the file provided in the inputfile. {utils.CGREEN}InCli -logs -inputfile "path to file/xxxx.log"{utils.CEND} 

        {utils.CYELLOW}-logs -store {utils.CEND}--> parses the files saved locally in order of creation(download) {utils.CGREEN}InCli -logs -store {utils.CEND} 
            {utils.CYELLOW}-error {utils.CEND}--> parses the files saved locally and displays only the ones with errors. {utils.CGREEN}InCli -logs -store -error {utils.CEND} 

        {utils.CYELLOW}-logs -folder path {utils.CEND}--> parses the files saved locally in order of creation(download) {utils.CGREEN}InCli -logs -folder /a/b/.../ {utils.CEND} {utils.CBLINK}NEW{utils.CEND}

        {utils.CYELLOW}-logs -last X {utils.CEND}--> parses the last X logs. X as number. {utils.CGREEN}InCli -u orgAlias -logs -last 10{utils.CEND}. Parses the last 10 logs in the org.
            {utils.CYELLOW}-loguser field:value{utils.CEND}, filter the logs for the specified user. The user can be specified by any field in the User Object. 
                {utils.CGREEN}-loguser Id:0053O000000IHneQAG, -loguser "name:Onboarding Site Guest User", -loguser Alias:John.Doe, -loguser FirstName:Onboarding, -loguser ProfileId:00e3O000000IHneQAG{utils.CEND}
            {utils.CYELLOW}-where X {utils.CEND}--> the where clause for a query on the ApexLog object.  
                {utils.CGREEN}InCli -logs -loguser Alias:John.Doe -where="Status<>'Success" -last 10" {utils.CEND} --> parses the last 10 logs for user John.Doe where the status is not Success
            {utils.CYELLOW}-file {utils.CEND}--> creates 2 files for the parsed file, txt and html format. 
            {utils.CYELLOW}-limitinfo {utils.CEND}--> prints the times when the limits (CPU, SOQL) are parsed.    

        {utils.CYELLOW}-logs -tail {utils.CEND}--> activelly parsers new debug logs as they are created.
            {utils.CYELLOW}-deletelogs {utils.CEND}--> as log files are processed, they are deleted from the server. The local copy remains. if -loguser is not provided it will default to your User.
            {utils.CYELLOW}-auto {utils.CEND}--> automaticallyu configures the collection of logs for -loguser. If no -loguser provided defaults to your user. {utils.CBLINK}NEW{utils.CEND}
            {utils.CYELLOW}-loguser field:value{utils.CEND}, filter the logs for the specified user. The user can be specified by any field in the User Object. 
            {utils.CGREEN}InCli -u orgAlias -logs -tail -loguser username:user@acme.com -deletelogs -auto.{utils.CEND} This command will listen to the logs for the loguser specified and delete the logs from the org as they are parsed (local copy is kept)

        {utils.CYELLOW}-logs{utils.CEND} --> lists log records in the org. Default 50 lines.  {utils.CGREEN}InCli -u orgAlias -logs {utils.CEND} 
            {utils.CYELLOW}-auto{utils.CEND}, refreshes the view every couple of seconds.     {utils.CGREEN}InCli -u orgAlias -logs -auto {utils.CEND}  {utils.CBLINK}NEW{utils.CEND}
            {utils.CYELLOW}-limit X{utils.CEND}, where X specifies the number of logs to list. Default 50 max 50K {utils.CGREEN}InCli -u orgAlias -logs -auto -limit 10{utils.CEND} """
        

    if '-hu' in args: return

    if '-h' in args:
        return help

    logId = getParam(args,'-logs')    
    last = getParam(args,'-last')  
    limit = getParam(args,'-limit')    
    limitInfo = True if '-limitinfo' in args else False   
    auto = True if '-auto' in args else False   
    folder = getParam(args,'-folder')    
    is_folder  = True if '-folder' in args else False   

    level = getParam(args,'-level') 
    loguser = getParam(args,'-loguser') 
    whereClause = getParam(args,'-where') 
    toFile = True if '-file' in args else False
    inputfile = getParam(args,'-inputfile') 
    tail = True if '-tail' in args else False
    deletelogs = True if '-deletelogs' in args else False
    store = True if '-store' in args else False
    store_logId = getParam(args,'-store') 
    error = True if '-error' in args else False

    if loguser == None:
        loguser = restClient.getConfigVar('loguser')

    if logId == None and last== None and inputfile==None and tail == False and store==False and is_folder==False:
        connectionInit(args)
        lim = 50 if limit == None else limit

        if auto:
            while True:
                debugLogs.printLogRecords(loguser=loguser,limit=lim,whereClause=whereClause)
                time.sleep(5)

        else:
            return debugLogs.printLogRecords(loguser=loguser,limit=lim,whereClause=whereClause)

    parseContext = {
        'logId' :logId,
        'filepath':inputfile,
        'printLimits':limitInfo,
        'lastN':last,
        'loguser':loguser,
        'level':level,
        'whereClause':whereClause,
        'writeToFile':toFile,
        'tail':tail,
        'deleteLogs':deletelogs,
        'parseStore':store,
        'logToStore':True,
        'print_only_errors':error,
        'operation':None,
        'auto':auto,
        'store_logId':store_logId,
        'search_dir':folder
    }

    if tail:
        connection = connectionInit(args)
        parseContext['connection'] = connection
        return debugLogs.do_parse_tail(parseContext)

    elif store:
        return debugLogs.do_parse_storage(parseContext)

    elif is_folder:
        return debugLogs.do_parse_storage(parseContext,search_dir=folder)

    elif inputfile != None:
        return debugLogs.do_parse_from_file(parseContext)

    elif logId != None:  #check if already downloaded        
        connectionInit(args)
        return debugLogs.do_parse_logId(parseContext)
    else:
        connectionInit(args)
        return debugLogs.do_parse_logs_lastN(parseContext)

def option_h(args,type="-h"):

    module = __import__('InCli')

    funcs = getmembers(sys.modules[__name__], isfunction)
    functions = [func[0] for func in funcs if func[0].startswith('option_')]

    for f in functions:
        if f == 'option_h':      continue
        args = type
        p = eval(f'{f}(args)')
        if p != None: print(p)

    print()
    print()

def option_hu(args):
    if '-h' in args:
        help = f"""
        {utils.CYELLOW}-hu{utils.CEND} --> additional help for utilities. """
        return help    

    option_h(args,"-hu")

def connectionInit(argsIn):

    userName_or_ofgAlias = getParam(argsIn,'-u') 
    return restClient.init(userName_or_ofgAlias)

def checkVersion():
    try:
        from importlib import metadata
    except ImportError:
        import importlib_metadata as metadata

    fileversion = metadata.version("InCli")

    try:
        restClient.init('ConnectionLess')
       # restClient.glog().info("Checking version")
        res = restClient.requestRaw(url='https://pypi.org/pypi/InCli/json')
        if res == 'No Response':
            print("No connection.")
            return
        version = res['urls'][1]['filename'].split('-')[1].split('.tar')[0]

        if version != fileversion:
            print()
            print(f"Version {fileversion}. {colorama.Fore.WHITE+colorama.Back.RED}Version {version} is available --> pip install InCli --upgrade{utils.CEND}")
        else:
            print()
            print(f"{utils.CFAINT}InCli version {fileversion} {utils.CEND}")
    except Exception as e:
        print(e)

def _main(argsIn):

    if "-debug" in argsIn:    restClient.setLoggingLevel(level=logging.DEBUG)
    else:  restClient.setLoggingLevel(logging.INFO)

    restClient.glog().debug("In debug mode")
    checkVersion()


    if '-h' in argsIn:
        colorama.just_fix_windows_console()

        help = f"""


        {utils.CYELLOW}-readme{utils.CEND} --> new functionality added and some example. 

        {utils.CYELLOW}-u{utils.CEND} --> username or org alias to be used to log into the org. Used in most commands. {utils.CGREEN}InCli -u collote@acme.com ... {utils.CEND} or {utils.CGREEN}InCli -u orgAlias... {utils.CEND} """
        print(help)

    funcs = getmembers(sys.modules[__name__], isfunction)
    functions = [func[0] for func in funcs if func[0].startswith('option_')]

    args = []
    for argv in argsIn:
        if argv == '|':     break
        args.append(argv)

    for arg in args:
        ar = arg.split(':')[0][1:]
        ar = f"option_{ar}"
        if ar in functions:
            res = None
            res = eval(f'{ar}(args)')
            if '-h' not in argsIn:
                return res

    if '-h' in argsIn:
        print()
        print(utils.CBOLD+"SFDX Commands:"+utils.CEND)
        print("InCli connects to the orgs as authorized in SFDX. Some useful commands are:")
        print()
        print(f" - {utils.CYELLOW}sfdx force:org:list --verbose --all {utils.CEND}--> to list all authorized Orgs and Connection Status")
        print(f" - {utils.CYELLOW}sfdx auth:web:login -r 'Instance Url' -a 'Alias' {utils.CEND}--> to re-authorize")
        print()

def main():
    argsIn = sys.argv

    try:
        _main(argsIn)

    except KeyboardInterrupt as e:
        print()
        print("ate ja")
    except Exception as e:
        colorama.just_fix_windows_console()

        utils.printException(e)
       # print(traceback.format_exc())

        if "-debug" in sys.argv:
            print(traceback.format_exc())

    print()
if __name__ == '__main__':
    main()




from . import query

def pricebookId(pricelistName,field='Name'):
    pb2Id = query.queryField(f" select vlocity_cmt__Pricebook2Id__c from vlocity_cmt__PriceList__c where {field} = '{pricelistName}' ")

    return pb2Id

def pricebookEntryId(pricebook2Id,productCode):
    pricebookEntryId = query.queryField(f" select Id from PricebookEntry where ProductCode='{productCode}' and Pricebook2Id = '{pricebook2Id}' ")

    return pricebookEntryId

def pricebookEntryId_pl(pricelistName,productCode,pricelistField='Name'):
    pb2Id = pricebookId(pricelistName,field=pricelistField)
    pbeId = pricebookEntryId(pb2Id,productCode)
    return pbeId







from . import query,utils
import simplejson,sys

def get_errors(where=None,limit=None):
    l = f" limit {limit} " if limit!= None else " limit 50 "
    q = f"""select Id,
                    OwnerId,
                    IsDeleted,
                    Name,
                    CreatedDate,
                    CreatedById,
                    LastModifiedDate,
                    SystemModstamp,
                    vlocity_cmt__ActionContainerComponent__c,
                    vlocity_cmt__TrackingService__c,
                    vlocity_cmt__VlocityInteractionToken__c,
                    vlocity_cmt__ErrorMessage__c,
                    vlocity_cmt__Timestamp__c,
                    vlocity_cmt__RequestPayload__c
                    from vlocity_cmt__VlocityTrackingEntry__c order by LastModifiedDate desc  {l}"""

    return query.query(q)


def print_error_list(limit=None):
    res = get_errors(limit=limit)
    if len(res['records'])==0: return print(f"There are no records to pring.")

    for record in res['records']:
        record['LastModifiedDate'] = record['LastModifiedDate'][0:19]
        if 'vlocity_cmt__ErrorMessage__c' in record and record['vlocity_cmt__ErrorMessage__c'] != None:
            record['Error'] = "Exception"
        record['LastModifiedDate'] = record['LastModifiedDate'][0:19]

    print(f"Printing {len(res['records'])} records from vlocity_cmt__VlocityTrackingEntry__c.")
    utils.printFormated(res['records'],"Id:LastModifiedDate:vlocity_cmt__ActionContainerComponent__c:vlocity_cmt__TrackingService__c:Name:Error",rename="vlocity_cmt__ActionContainerComponent__c%acc:LastModifiedDate%Time")

def _get_record_error_message(record):
    input_data = simplejson.loads(record['vlocity_cmt__InputData__c'])

    error_message = 'Other'

    if record['vlocity_cmt__SourceType__c'] == 'Odin':
        error_message = input_data['ErrorMessage__c'][0:70]

    elif record['vlocity_cmt__SourceType__c'] == 'Integration Procedure':
        if  'ErrorMessage'  in input_data:
            if 'ErrorCode' in input_data:
                error_message  = f"{input_data['ErrorCode']} {input_data['ErrorMessage']} "
            else:
                error_message  = f"{input_data['ErrorMessage']} "

        elif 'GenericLogMessage' in input_data:
            error_message = input_data['GenericLogMessage'][0:70]
        else: 
            if 'StepResult' in input_data:
                if 'result' in input_data['StepResult']:
                    if 'HTTPStatusCode' in input_data['StepResult']['result']:
                        error_message  = f"HttpCode {input_data['StepResult']['result']['HTTPStatusCode']} "
                    if 'errors' in input_data['StepResult']['result']:
                        error_message  = str(input_data['StepResult']['result']['errors'])

                if 'info' in input_data['StepResult']:
                    if 'status' in input_data['StepResult']['info']:
                        error_message  = f"{input_data['StepResult']['info']['status']} {input_data['StepResult']['info']['statusCode']}"
                if 'error' in input_data['StepResult']:
                        error_message  = f"{input_data['StepResult']['error']}"
    return error_message

def print_error_records(last=None):
    limit = last if last!=None else 50
    res = get_errors(limit=limit)

    for record in res['records']:
        record['vlocity_cmt__ErrorTime__c'] = record['vlocity_cmt__ErrorTime__c'][0:19]
        record['ErrorMessage__c'] = _get_record_error_message(record)

        _print_error_record(record,print_renames=False)
        print()
        print()

   # utils.printFormated(res['records'],"Id:vlocity_cmt__ErrorTime__c:vlocity_cmt__SourceType__c:vlocity_cmt__SourceName__c:vlocity_cmt__Action__c:ErrorMessage__c",rename="vlocity_cmt__SourceType__c%sourceType:vlocity_cmt__ObjectName__c%ObjectName:vlocity_cmt__ErrorTime__c%time")

def query_and_print_error_record(id,tofile=False):
    q = f"select fields(all) from vlocity_cmt__VlocityTrackingEntry__c where Id='{id}'"
    records = query.queryRecords(q)

    if len(records)>0:
        record = records[0]
        _print_error_record(record,tofile=tofile)
    else:
        print(f"Record Id {id} not found.")

def _print_error_record(record,print_renames=True,tofile=False):
  #  record['ErrorMessage__c'] = _get_record_error_message(record)
    record['LastModifiedDate'] = record['LastModifiedDate'][0:19]

    utils.printFormated(record,"Id:vlocity_cmt__ErrorTime__c:vlocity_cmt__SourceType__c:vlocity_cmt__SourceName__c:vlocity_cmt__Action__c:ErrorMessage__c",rename="vlocity_cmt__SourceType__c%sourceType:vlocity_cmt__ObjectName__c%ObjectName:vlocity_cmt__ErrorTime__c%time",print_renames=print_renames)

    for key in list(record.keys()):
        if record[key] == None:
            record.pop(key)

    if 'vlocity_cmt__ErrorMessage__c' in record:
        record['vlocity_cmt__ErrorMessage__c'] = simplejson.loads(str(record['vlocity_cmt__ErrorMessage__c']))
        exception = record['vlocity_cmt__ErrorMessage__c']['Exception']
        if 'Description' in exception:
            exception['Description'] = simplejson.loads(exception['Description'])
   
    if 'vlocity_cmt__RequestPayload__c' in record:
        record['vlocity_cmt__RequestPayload__c'] = simplejson.loads(record['vlocity_cmt__RequestPayload__c'])

    if 'vlocity_cmt__ResponsePayload__c' in record:
        record['vlocity_cmt__ResponsePayload__c'] = simplejson.loads(record['vlocity_cmt__ResponsePayload__c'])

    if 'vlocity_cmt__Data__c' in record:
        record['vlocity_cmt__Data__c'] = simplejson.loads(record['vlocity_cmt__Data__c'])

    json_formatted_str = simplejson.dumps(record, indent=2, ensure_ascii=False)
  #  print(json_formatted_str)

    if tofile == True:
        original_stdout = sys.stdout
        filename = f"{record['Id']}.json"
        with open(filename, 'w') as f:
            sys.stdout = f 
            print(json_formatted_str)
            sys.stdout = original_stdout 
            print()
            print(f"   File {filename} created.")
    else:  
        print(json_formatted_str)



import simplejson as json
import os

from . import file,utils

def writeFileinFolder(folder,filename,data):
    return write(f"{folder}/{filename}",data)

def write(filepath,data):
    filepath = file.addExtension(filepath,'json')
    str = json.dumps(data, indent=2)
    return file.write(filepath,str,'w')

def readFolder(folder,filename):
    read(f"{folder}/{filename}")

def read(filepath):
    filepath = file.addExtension(filepath,'json')
    str = file.read(filepath)
   # obj = None
    try:
        obj= json.loads(str)
    
    except ValueError as e:
        extra = e.args[0] if len(e.args)>0 else ''
        utils.raiseException(f'JSON_PARSING:{type(e)}',f"{e.msg} {extra}",other=f"File {filepath} seems to be a malformed JSON?")
        #utils.printException(e)
    return obj

#looks for a field targetKey with value targeValue and optionally ¡n the same object/dictionary replaces the targetKey with the targetValue
def parse(json, targetKey='', targetKeyValue='',replaceKey='', replaceValue='',printProducts=False):

    #if it is not a dictionary or an array, its a value so return
    if isinstance(json,dict)==False and isinstance(json,list)==False:
        return ''  #continue

    for key in json.keys():
        if key == 'PricebookEntry':  #breaks it
            continue

        if key == targetKey:
            value = json[key]
            if 'value' in json[key]:  #sometime the value is in a nested filed called value
                value = json[key]['value']
            if value == targetKeyValue:
                if replaceKey!='':
                    json[replaceKey] = replaceValue
                return json     

        if isinstance(json[key],dict):
            ret = parse(json[key],targetKey, targetKeyValue,replaceKey,replaceValue)
            if ret != '':
                return ret
            ident = ident-1

        if isinstance(json[key],list):
            for l in json[key]:
                ret = parse(l,targetKey,targetKeyValue,replaceKey,replaceValue)
                if ret != '':
                    return ret
                ident = ident-1
    return ''
    
def getChild(json,name):
    for key in json.keys():
        if key == name:
            return json[key]
    return ''
